#
# Copyright (c) 2002, Art Haas
#
# This file is part of PythonCAD.
# 
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

#
# define __all__ so that "from Interface/Gtk import *" works
#

__all__ = [
    'gtkconobjs',
    'gtkdimension',
    'gtkdimprefs',
    'gtkedit',
    'gtkentities',
    'gtkimage',
    'gtklayers',
    'gtkmenus',
    'gtkmodify',
    'gtkprefs'
    ]
