#
# Copyright (c) 2002, 2003, 2004 Art Haas
#
# This file is part of PythonCAD.
#
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# construction line/circle base class
#
# These variables provide the defaults for
# the construction line style attributes
#

import graphicobject
import style
import linetype
import color
import tolerance
import point

class ConstructionObject(graphicobject.GraphicObject):
    """A base class for construction lines and circles.

This class is meant to provide the most basic bits for
construction lines and circles. All construction lines
and circles will share a common Style object style, meaning
all instances will be drawn with the same linetype, have
the same color, and be the same thickness. Construction
entities should never be plotted out, however.
    """

    # static class variables

    __defstyle = style.Style(u'Default Construction Object Style',
                             linetype.Linetype(u'Construction Line', [2,2]),
                             color.Color(255, 0, 0),
                             0.0)

    def __init__(self, **kw):
        graphicobject.GraphicObject.__init__(self, ConstructionObject.__defstyle, **kw)

    def setStyle(self, s):
        pass

    def setColor(self, c):
        pass

    def setLinetype(self, l):
        pass

    def setThickness(self, t):
        pass
