#!/usr/bin/env python
#
# Copyright (c) 2002, 2003, 2004 Art Haas
#

#
# Install pythoncad using the distutils method
#

from distutils.core import setup

setup(name="PythonCAD",
      version="DS1-R16",
      description="CAD built from Python",
      author="Art Haas",
      author_email="ahaas@airmail.net",
      url="http://www.pythoncad.org/",
      license="GPL",
      extra_path = "PythonCAD",
      packages=['Generic',
                'Interface',
                'Interface/Cocoa',
                'Interface/Gtk'],
     )
